import dataclasses
import inspect
import typing

from rest_framework.utils.model_meta import RelationInfo

from rest_framework_dataclasses import typing_utils

T = typing.TypeVar('T')


@dataclasses.dataclass
class DataclassDefinition:
    dataclass_type: type
    fields: typing.Dict[str, dataclasses.Field]
    field_types: typing.Dict[str, type]


@dataclasses.dataclass
class TypeInfo:
    is_many: bool
    is_mapping: bool
    is_final: bool
    is_nullable: bool
    base_type: type


def get_dataclass_definition(dataclass_type: type) -> DataclassDefinition:
    """
    Given a dataclass class, returns a dictionary mapping field names to
    `dataclasses.Field` instances describing all fields on the dataclass.
    """
    # Resolve the typehint from the dataclass fields (which can be stringified, especially with PEP 563 nowadays) to
    # actual type objects. Based on the discussion in https://stackoverflow.com/a/55938344.
    types = typing.get_type_hints(dataclass_type)

    # Disable PyCharm warning here, as it is wrong.
    # noinspection PyDataclass
    fields = {field.name: field for field in dataclasses.fields(dataclass_type)}

    return DataclassDefinition(dataclass_type, fields, types)


def get_type_info(tp: type) -> TypeInfo:
    """
    Reduce iterable and optional types to their 'base' types.
    """
    is_final = typing_utils.is_final_type(tp)
    if is_final:
        tp = typing_utils.get_final_type(tp)
        if tp is typing.Any:
            # This used to be supported, but Python 3.10 raises in get_type_hints() if it encounters a plain Final hint.
            raise TypeError('Plain typing.Final is not valid as a type argument.')

    is_nullable = typing_utils.is_optional_type(tp)
    if is_nullable:
        tp = typing_utils.get_optional_type(tp)

    is_mapping = typing_utils.is_mapping_type(tp)
    is_many = typing_utils.is_iterable_type(tp)

    if is_mapping:
        tp = typing_utils.get_mapping_value_type(tp)
    elif is_many:  # This must be elif (instead of if), as otherwise we'd reduce mappings twice as they're also iterable
        tp = typing_utils.get_iterable_element_type(tp)

    if typing_utils.is_type_variable(tp):
        tp = typing_utils.get_variable_type_substitute(tp)

    if tp is typing.Any:
        tp = None

    return TypeInfo(is_many, is_mapping, is_final, is_nullable, tp)


def get_relation_info(type_info: TypeInfo) -> RelationInfo:
    """
    Given a type_info that references a Model, extract the RelationInfo.
    """
    return RelationInfo(
        # there's no foreign key field
        model_field=None,
        related_model=type_info.base_type,
        to_many=type_info.is_many,
        # as there's no foreign key, it also cannot reference a field on the referenced model
        to_field=None,
        has_through_model=False,
        # we're never the model
        reverse=False
    )


def lookup_type_in_mapping(mapping: typing.Dict[type, T], key: type) -> T:
    try:
        # _SpecialForm types like Literal, NoReturn don't have an __mro__ attribute
        bases = inspect.getmro(key)
    except AttributeError:
        raise KeyError("Special type {typ} not supported.".format(typ=key))

    for cls in bases:
        if cls in mapping:
            return mapping[cls]

    raise KeyError("Class '{cls}' not found in lookup.".format(cls=key))



import enum

from unittest import TestCase

from rest_framework.exceptions import ValidationError

from rest_framework_dataclasses.fields import DefaultDecimalField, EnumField


class FieldTest(TestCase):
    def test_decimal_field(self):
        # check no parameters are required (this shouldn't throw)
        field = DefaultDecimalField()

        # check no parameters are overwritten
        field = DefaultDecimalField(max_digits=4, decimal_places=6)
        self.assertEqual(field.max_digits, 4)
        self.assertEqual(field.decimal_places, 6)

    def test_enum_field(self):
        class Color(enum.Enum):
            RED = 'FF0000'
            GREEN = '00FF00'
            BLUE = '0000FF'

        field = EnumField(Color)
        self.assertDictEqual(field.choices, {'FF0000': 'RED', '00FF00': 'GREEN', '0000FF': 'BLUE'})
        self.assertEqual(field.to_representation(Color.GREEN), '00FF00')
        self.assertEqual(field.to_internal_value('00FF00'), Color.GREEN)
        with self.assertRaises(ValidationError):
            field.to_internal_value('RED')

        field = EnumField(Color, by_name=True)
        self.assertDictEqual(field.choices, {'RED': 'RED', 'GREEN': 'GREEN', 'BLUE': 'BLUE'})
        self.assertEqual(field.to_internal_value('GREEN'), Color.GREEN)
        self.assertEqual(field.to_representation(Color.GREEN), 'GREEN')
        with self.assertRaises(ValidationError):
            field.to_internal_value('FF0000')

        self.assertEqual(field.to_representation('RED'), 'RED')
        with self.assertRaises(ValidationError):
            field.to_representation('FFFFFF')



from rest_framework.fields import DecimalField, ChoiceField


class DefaultDecimalField(DecimalField):
    def __init__(self, **kwargs):
        if 'max_digits' not in kwargs:
            kwargs['max_digits'] = None
        if 'decimal_places' not in kwargs:
            # Maybe this should be configurable, but it doesn't seem that useful. File an issue if you want it to.
            kwargs['decimal_places'] = 2

        super(DefaultDecimalField, self).__init__(**kwargs)


class EnumField(ChoiceField):
    def __init__(self, enum_class, by_name=False, **kwargs):
        self.enum_class = enum_class
        self.by_name = by_name
        if 'choices' not in kwargs:
            kwargs['choices'] = [(self.to_representation(member), member.name) for member in self.enum_class]

        super(EnumField, self).__init__(**kwargs)

    def to_internal_value(self, data):
        try:
            if self.by_name:
                return self.enum_class[data]
            else:
                return self.enum_class(data)
        except (KeyError, ValueError):
            self.fail('invalid_choice', input=data)

    def to_representation(self, value):
        # Some external libraries expect to be able to call to_representation() with the key from the choices
        # array, which seems at least somewhat reasonable. See #40.
        if not isinstance(value, self.enum_class):
            if value in self.choices:
                return value
            self.fail('invalid_choice', input=value)

        if self.by_name:
            return value.name
        else:
            return value.value



import copy
import dataclasses
import datetime
import decimal
import enum
import typing
import uuid

from unittest import TestCase

from rest_framework import fields

from rest_framework_dataclasses.serializers import DataclassSerializer
from rest_framework_dataclasses.types import Literal


class Gender(enum.Enum):
    MALE = 'male'
    FEMALE = 'female'
    OTHER = 'other'


@dataclasses.dataclass
class Pet:
    animal: Literal['cat', 'dog']
    name: str
    weight: typing.Optional[decimal.Decimal] = \
        dataclasses.field(default=None, metadata={'serializer_kwargs': {'max_digits': 4, 'decimal_places': 1}})


class PetSerializer(DataclassSerializer):
    class Meta:
        dataclass = Pet


@dataclasses.dataclass
class Person:
    id: uuid.UUID
    name: str
    email: str
    phone: typing.List[str]
    gender: typing.Optional[Gender] = None
    length: typing.Optional[decimal.Decimal] = None
    pets: typing.Optional[typing.List[Pet]] = None
    birth_date: typing.Optional[datetime.date] = None
    favorite_pet: typing.Optional[Pet] = \
        dataclasses.field(default=None, metadata={'serializer_field': PetSerializer(allow_null=True)})
    movie_ratings: typing.Optional[typing.Dict[str, int]] = None

    def age(self) -> int:
        # don't use current date to keep test deterministic, and we don't care that this calculation is actually wrong.
        return datetime.date(2020, 1, 1).year - self.birth_date.year if self.birth_date else None


class PersonSerializer(DataclassSerializer):
    full_name = fields.CharField(source='name')
    email = fields.EmailField()
    slug = fields.SlugField(source='name', read_only=True)

    class Meta:
        dataclass = Person
        fields = ('id', 'full_name', 'email', 'phone', 'gender', 'length', 'pets', 'birth_date', 'favorite_pet',
                  'movie_ratings', 'slug', 'age')
        extra_kwargs = {
            'id': {'format': 'hex'},
            'phone': {'child_kwargs': {'max_length': 15}},
            'pets': {'child_kwargs': {'extra_kwargs': {'weight': {'max_digits': 4, 'decimal_places': 1}}}},
        }


# noinspection PyUnresolvedReferences
class FunctionalTestMixin:
    def test_serialize(self):
        serializer = self.serializer(instance=self.instance)
        self.assertDictEqual(serializer.data, {**self.representation, **self.representation_readonly})

    def test_validated_data(self):
        serializer = self.serializer(data=self.representation)
        serializer.is_valid(raise_exception=True)

        self.assertEqual(serializer.validated_data, self.instance)

    def test_create(self: TestCase):
        serializer = self.serializer(data=self.representation)
        serializer.is_valid(raise_exception=True)
        instance = serializer.save()

        self.assertEqual(instance, self.instance)

    def test_update(self: TestCase):
        empty_instance = copy.deepcopy(self.instance)
        for field in dataclasses.fields(empty_instance):
            setattr(empty_instance, field.name, None)

        serializer = self.serializer(instance=empty_instance, data=self.representation)
        serializer.is_valid(raise_exception=True)
        instance = serializer.save()

        self.assertIs(instance, empty_instance)
        self.assertEqual(instance, self.instance)


class PetTest(TestCase, FunctionalTestMixin):
    serializer = PetSerializer
    instance = Pet(animal='cat', name='Milo')
    representation = {'animal': 'cat', 'name': 'Milo', 'weight': None}
    representation_readonly = {}


class PersonTest(TestCase, FunctionalTestMixin):
    maxDiff = None
    serializer = PersonSerializer
    instance = Person(
        id=uuid.UUID('28ee3ae5-480b-46bd-9ae4-c61cf8341b95'),
        name='Alice',
        email='alice@example.com',
        length=decimal.Decimal('1.68'),
        phone=['+31-6-1234-5678', '+31-20-123-4567'],
        gender=Gender.FEMALE,
        pets=[Pet(animal='cat', name='Milo', weight=decimal.Decimal('10.8')),
              Pet(animal='dog', name='Max', weight=decimal.Decimal('123.4'))],
        birth_date=datetime.date(1980, 4, 1),
        favorite_pet=Pet(animal='cat', name='Luna', weight=None),
        movie_ratings={'Star Wars': 8, 'Titanic': 4}
    )
    representation = {
        'id': '28ee3ae5480b46bd9ae4c61cf8341b95',
        'full_name': 'Alice',
        'email': 'alice@example.com',
        'length': '1.68',
        'phone': ['+31-6-1234-5678', '+31-20-123-4567'],
        'gender': 'female',
        'pets': [
            {'animal': 'cat', 'name': 'Milo', 'weight': '10.8'},
            {'animal': 'dog', 'name': 'Max', 'weight': '123.4'}
        ],
        'birth_date': '1980-04-01',
        'age': 40,
        'favorite_pet': {'animal': 'cat', 'name': 'Luna', 'weight': None},
        'movie_ratings': {'Star Wars': 8, 'Titanic': 4},
    }
    representation_readonly = {
        'slug': 'Alice'
    }


class EmptyPersonTest(TestCase, FunctionalTestMixin):
    serializer = PersonSerializer
    instance = Person(
        id=uuid.UUID('28ee3ae5-480b-46bd-9ae4-c61cf8341b95'),
        name='Alice',
        email='alice@example.com',
        phone=[],
    )
    representation = {
        'id': '28ee3ae5480b46bd9ae4c61cf8341b95',
        'full_name': 'Alice',
        'email': 'alice@example.com',
        'phone': [],
        'favorite_pet': None,
    }
    representation_readonly = {
        'slug': 'Alice',
        'length': None,
        'gender': None,
        'pets': None,
        'birth_date': None,
        'age': None,
        'movie_ratings': None
    }


class PartialPersonTest(TestCase):
    def test_update(self):
        input_instance = Person(
            id=uuid.UUID('28ee3ae5-480b-46bd-9ae4-c61cf8341b95'),
            name='Alice',
            email='alice@example.com',
            phone=['+31-6-1234-5678', '+31-20-123-4567'],
            favorite_pet=Pet('cat', 'Luna'),
            pets=[Pet('dog', 'Bella'), Pet('cat', 'Luna')],
        )

        partial_representation = {
            'full_name': 'Bob',
            'email': 'bob@example.com',
            'favorite_pet': {'name': 'Molly'},
            'pets': [{'animal': 'cat', 'name': 'Molly'}],
        }

        expected_output = dataclasses.replace(input_instance,
                                              name='Bob',
                                              email='bob@example.com',
                                              favorite_pet=Pet(animal='cat', name='Molly'),
                                              pets=[Pet('cat', 'Molly')])

        serializer = PersonSerializer(instance=input_instance, data=partial_representation, partial=True)
        serializer.is_valid(raise_exception=True)
        output_instance = serializer.save()

        self.assertIs(output_instance, input_instance)
        self.assertEqual(output_instance, expected_output)



import copy
import dataclasses
import datetime
import typing

from unittest import TestCase

from django.core.exceptions import ImproperlyConfigured
from rest_framework import fields, serializers
from rest_framework.fields import empty

from rest_framework_dataclasses import types
from rest_framework_dataclasses.serializers import (_strip_empty_sentinels,
                                                    DataclassSerializer, HyperlinkedDataclassSerializer)


@dataclasses.dataclass
class Person:
    name: str
    length: int
    birth_date: typing.Optional[datetime.date] = None

    def age(self) -> int:
        pass


@dataclasses.dataclass
class Group:
    leader: Person
    people: typing.List[Person]


class SerializerTest(TestCase):
    def create_serializer(self, dataclass=None, arguments=None, declared=None, meta=None) -> DataclassSerializer:
        arguments = arguments or {}
        classdict = declared or {}

        if meta is not None:
            if dataclass:
                meta['dataclass'] = dataclass
            classdict['Meta'] = type('Meta', (), meta)
        elif dataclass:
            arguments['dataclass'] = dataclass

        serializer_type = type('TestSerializer', (DataclassSerializer, ), classdict)
        return serializer_type(**arguments)

    def test_definition(self):
        definition = self.create_serializer(Person).dataclass_definition
        self.assertIs(definition.dataclass_type, Person)
        self.assertIn('name', definition.fields)
        self.assertEqual(definition.field_types['name'], str)
        self.assertIn('length', definition.fields)
        self.assertEqual(definition.field_types['length'], int)
        self.assertIn('birth_date', definition.fields)
        self.assertEqual(definition.field_types['birth_date'], typing.Optional[datetime.date])

        # both `dataclass` and `Meta` specified
        with self.assertRaises(AssertionError):
            definition = self.create_serializer(arguments={'dataclass': Person},
                                                meta={}).dataclass_definition

        # no `dataclass` parameter and missing `Meta`
        with self.assertRaises(AssertionError):
            definition = self.create_serializer().dataclass_definition

        # no `dataclass` parameter and invalid `Meta`
        with self.assertRaises(AssertionError):
            definition = self.create_serializer(meta={}).dataclass_definition

        # non-dataclass
        with self.assertRaises(ValueError):
            definition = self.create_serializer(str).dataclass_definition

    def test_strip_empty(self):
        # simple
        in_data = Person(name='Alice', length=123, birth_date=empty)
        out_data = Person(name='Alice', length=123, birth_date=None)
        self.assertEqual(_strip_empty_sentinels(in_data), out_data)

        # nested dataclasses
        in_data = Group(leader=Person(name='Alice', length=123, birth_date=empty),
                        people=[Person(name='Bob', length=456, birth_date=empty)])
        out_data = Group(leader=Person(name='Alice', length=123),
                         people=[Person(name='Bob', length=456)])
        self.assertEqual(_strip_empty_sentinels(in_data), out_data)

    def test_save(self):
        def mock_save(validated_data, instance=None, **kwargs):
            serializer = self.create_serializer(Person)
            serializer._errors = []
            serializer._validated_data = validated_data
            if instance:
                serializer.instance = instance
            return serializer.save(**kwargs)

        # regular create
        in_data = Person(name='Alice', length=123)
        out_data = mock_save(in_data)
        self.assertEqual(in_data, out_data)

        # regular update
        inst = Person(name='Alice', length=123)
        in_data = dataclasses.replace(inst, length=456)
        out_data = mock_save(in_data, instance=inst)
        self.assertIs(out_data, inst)
        self.assertEqual(out_data, in_data)

        # create with kwargs
        in_data = Person(name='Alice', length=123)
        out_data = mock_save(in_data, length=456)
        self.assertEqual(out_data, dataclasses.replace(in_data, length=456))

        # update with kwargs
        inst = Person(name='Bob', length=789)
        in_data = Person(name='Alice', length=456)
        out_data = mock_save(in_data, instance=inst, length=456)
        self.assertIs(out_data, inst)
        self.assertEqual(out_data, dataclasses.replace(in_data, length=456))

        # full update with explicit reset of field to default value
        inst = Person(name='Alice', length=123, birth_date=datetime.datetime(2020, 2, 2))
        in_data = Person(name='Alice', length=123, birth_date=None)
        out_data = mock_save(in_data, instance=inst)
        self.assertIs(out_data, inst)
        self.assertEqual(out_data.birth_date, None)

        # full update with unsupplied optional field
        inst = Person(name='Alice', length=123, birth_date=datetime.datetime(2020, 2, 2))
        in_data = Person(name='Alice', length=123, birth_date=empty)
        out_data = mock_save(in_data, instance=inst)
        self.assertIs(out_data, inst)
        self.assertEqual(out_data.birth_date, datetime.datetime(2020, 2, 2))

        # partial update with missing required field
        inst = Person(name='Alice', length=123)
        in_data = Person(name='Bob', length=empty)
        out_data = mock_save(in_data, instance=inst)
        self.assertIs(out_data, inst)
        self.assertEqual(out_data.name, 'Bob')
        self.assertEqual(out_data.length, 123)

        # not validated
        with self.assertRaises(AssertionError):
            self.create_serializer(Person).save()

    def test_nested_save(self):
        def check(dataclass, representation, instance):
            empty_instance = copy.deepcopy(instance)
            for field in dataclasses.fields(empty_instance):
                setattr(empty_instance, field.name, None)

            serializer = self.create_serializer(dataclass, arguments={'data': representation})
            self.assertTrue(serializer.is_valid(raise_exception=False))
            self.assertEqual(serializer.save(), instance)

            serializer = self.create_serializer(dataclass, arguments={'data': representation,
                                                                      'instance': empty_instance})
            self.assertTrue(serializer.is_valid(raise_exception=False))
            self.assertEqual(serializer.save(), instance)

        # simple dataclass with a single field
        simple = dataclasses.make_dataclass('child', [('value', str)])
        check(simple, {'value': 'A'}, simple('A'))

        # nested dataclass
        parent = dataclasses.make_dataclass('parent', [('field', simple)])
        check(parent, {'field': {'value': 'A'}}, parent(simple('A')))

        # a nested dataclass that's optional
        optional = dataclasses.make_dataclass('optional', [('field', typing.Optional[simple])])
        check(optional, {'field': {'value': 'A'}}, optional(simple('A')))
        check(optional, {'field': None}, optional(None))

        # we check all possible kinds of (optionally) nested dataclasses in lists and dictionaries here, as historically
        # this code has given us quite some problems (#13, #15).
        # a list of nested dataclasses
        listvalue = dataclasses.make_dataclass('listvalue',
                                               [('field', typing.Iterable[simple])])
        check(listvalue, {'field': []}, listvalue([]))
        check(listvalue, {'field': [{'value': 'A'}]}, listvalue([simple('A')]))

        # an optional list of nested dataclasses
        optionallist = dataclasses.make_dataclass('optionallist',
                                                  [('field', typing.Optional[typing.List[simple]])])
        check(optionallist, {'field': None}, optionallist(None))
        check(optionallist, {'field': []}, optionallist([]))
        check(optionallist, {'field': [{'value': 'A'}]}, optionallist([simple('A')]))

        # a list of optional nested dataclasses
        listoptional = dataclasses.make_dataclass('listoptional',
                                                  [('field', typing.List[typing.Optional[simple]])])
        check(listoptional, {'field': []}, listoptional([]))
        check(listoptional, {'field': [None]}, listoptional([None]))
        check(listoptional, {'field': [{'value': 'A'}]}, listoptional([simple('A')]))

        # a dictionary of nested dataclasses
        dictvalue = dataclasses.make_dataclass('dictvalue',
                                               [('field', typing.Mapping[str, simple])])
        check(dictvalue, {'field': {}}, dictvalue({}))
        check(dictvalue, {'field': {'K': {'value': 'A'}}}, dictvalue({'K': simple('A')}))

        # an optional dictionary of nested dataclasses
        optionaldict = dataclasses.make_dataclass('optionaldict',
                                                  [('field', typing.Optional[typing.Dict[str, simple]])])
        check(optionaldict, {'field': None}, optionaldict(None))
        check(optionaldict, {'field': {}}, optionaldict({}))
        check(optionaldict, {'field': {'K': {'value': 'A'}}}, optionaldict({'K': simple('A')}))

        # a dictionary of optional nested dataclasses
        dictoptional = dataclasses.make_dataclass('dictoptional',
                                                  [('field', typing.Dict[str, typing.Optional[simple]])])
        check(dictoptional, {'field': {}}, dictoptional({}))
        check(dictoptional, {'field': {'K': None}}, dictoptional({'K': None}))
        check(dictoptional, {'field': {'K': {'value': 'A'}}}, dictoptional({'K': simple('A')}))

    def test_get_fields(self):
        @dataclasses.dataclass
        class TestPerson:
            name: str
            email: str = dataclasses.field(metadata={'serializer_field': fields.EmailField()})
            birth_date: typing.Optional[datetime.date] = None

        f = self.create_serializer(TestPerson,
                                   declared={'slug': fields.SlugField(source='name')},
                                   meta={'read_only_fields': ['birth_date'],
                                         'extra_kwargs': {'name': {'max_length': 15}}}).get_fields()

        self.assertEqual(len(f), 4)

        self.assertIsInstance(f['name'], fields.CharField)
        self.assertTrue(f['name'].required)
        self.assertEqual(f['name'].max_length, 15)

        self.assertIsInstance(f['email'], fields.EmailField)

        self.assertIsInstance(f['birth_date'], fields.DateField)
        self.assertTrue(f['birth_date'].read_only)

        self.assertIsInstance(f['slug'], fields.SlugField)
        self.assertEqual(f['slug'].source, 'name')

        with self.assertRaises(AssertionError):
            self.create_serializer(TestPerson,
                                   meta={'read_only_fields': ['email']}).get_fields()

        with self.assertRaises(AssertionError):
            self.create_serializer(TestPerson,
                                   declared={'slug': fields.SlugField(source='name')},
                                   meta={'read_only_fields': ['slug']}).get_fields()

    def test_get_field_names(self):
        def check(serializer, expected):
            self.assertListEqual(serializer.get_field_names(), expected)

        check(self.create_serializer(Person),
              ['name', 'length', 'birth_date'])
        check(self.create_serializer(Person, meta={'fields': ['name', 'length']}),
              ['name', 'length'])
        check(self.create_serializer(Person, meta={'fields': serializers.ALL_FIELDS}),
              ['name', 'length', 'birth_date'])
        check(self.create_serializer(Person, meta={'fields': (serializers.ALL_FIELDS, 'age')}),
              ['age', 'name', 'length', 'birth_date'])
        check(self.create_serializer(Person, meta={'fields': ('age', )}),
              ['age'])
        check(self.create_serializer(Person, meta={'exclude': ['name']}),
              ['length', 'birth_date'])
        check(self.create_serializer(Person, declared={'age': fields.ReadOnlyField()}),
              ['age', 'name', 'length', 'birth_date'])

        # invalid `fields` specification
        with self.assertRaises(TypeError):
            self.create_serializer(Person, meta={'fields': 'invalid'}).get_field_names()

        # invalid `exclude` specification
        with self.assertRaises(TypeError):
            self.create_serializer(Person, meta={'exclude': 'invalid'}).get_field_names()

        # declared field not in `fields`
        with self.assertRaises(AssertionError):
            self.create_serializer(Person,
                                   declared={'name': fields.SlugField()},
                                   meta={'fields': ['length']}).get_field_names()

        # declared field in `exclude`
        with self.assertRaises(AssertionError):
            self.create_serializer(Person,
                                   declared={'name': fields.SlugField()},
                                   meta={'exclude': ['name']}).get_field_names()

        # non-existing field in `fields`
        with self.assertRaises(AssertionError):
            self.create_serializer(Person, meta={'fields': ['nonexisting']}).get_field_names()

        # non-existing field in `exclude`
        with self.assertRaises(AssertionError):
            self.create_serializer(Person, meta={'exclude': ['nonexisting']}).get_field_names()

        # both `fields` and `exclude` specified
        with self.assertRaises(AssertionError):
            self.create_serializer(Person, meta={'fields': ['name'], 'exclude': ['length']}).get_field_names()

    def test_create_field(self):
        @dataclasses.dataclass
        class TestPerson:
            name: str
            length: int = dataclasses.field(metadata={'serializer_kwargs': {'max_value': 200}})
            species: types.Final[str] = 'Human'

            def age(self) -> int:
                pass

        serializer = self.create_serializer(TestPerson)

        # From a field
        name_field = serializer.create_field('name', {'max_length': 20, 'trim_whitespace': False})
        self.assertIsInstance(name_field, fields.CharField)
        self.assertEqual(name_field.max_length, 20)
        self.assertEqual(name_field.trim_whitespace, False)

        # Field with metadata arguments
        length_field = serializer.create_field('length', {})
        self.assertIsInstance(length_field, fields.IntegerField)
        self.assertEqual(length_field.max_value, 200)

        # Function field
        age_field = serializer.create_field('age', {})
        self.assertIsInstance(age_field, fields.ReadOnlyField)

        # Aliased field
        aliased_field = serializer.create_field('aliased', {'source': 'name'})
        self.assertIsInstance(aliased_field, fields.CharField)
        self.assertEqual(aliased_field.source, 'name')

        # Aliased to non-existing field
        with self.assertRaises(ImproperlyConfigured):
            unknown_field = serializer.create_field('aliased', {'source': 'nonexisting'})

        # Final field with a default value
        species_field = serializer.create_field('species', {})
        self.assertIsInstance(species_field, fields.CharField)
        self.assertTrue(species_field.read_only)

        # Plain Final annotation should raise
        with self.assertRaises(TypeError):
            @dataclasses.dataclass
            class TestFinal:
                species: types.Final = 'Human'

            self.create_serializer(TestFinal).create_field('species', {})

    def test_include_extra_kwargs(self):
        serializer = self.create_serializer()

        # Normal updates
        self.assertDictEqual(serializer.include_extra_kwargs({'a': 0}, {'a': 1}), {'a': 1})
        self.assertDictEqual(serializer.include_extra_kwargs({'a': 0}, {'b': 1}), {'a': 0, 'b': 1})

        # Special cases
        self.assertDictEqual(serializer.include_extra_kwargs({'min_length': 3}, {'read_only': True}),
                             {'read_only': True})
        self.assertDictEqual(serializer.include_extra_kwargs({'required': False}, {'default': 1}),
                             {'default': 1})
        self.assertDictEqual(serializer.include_extra_kwargs({'read_only': True}, {'required': True}),
                             {'read_only': True})
        self.assertDictEqual(serializer.include_extra_kwargs({}, {'child_kwargs': {}}),
                             {})

    def test_get_extra_kwargs(self):
        def get(extra_kwargs=None, read_only_fields=None):
            meta = {'extra_kwargs': extra_kwargs, 'read_only_fields': read_only_fields}
            return self.create_serializer(meta=meta).get_extra_kwargs()

        self.assertDictEqual(get(extra_kwargs={'name': {'max_length': 20}}),
                             {'name': {'max_length': 20}})
        self.assertDictEqual(get(extra_kwargs={'name': {'max_length': 20}}, read_only_fields=['name']),
                             {'name': {'max_length': 20, 'read_only': True}})

        with self.assertRaises(TypeError):
            get(read_only_fields='invalid')

        with self.assertRaises(AssertionError):
            self.create_serializer(meta={'readonly_fields': ['name']}).get_extra_kwargs()

    def test_to_internal_value(self):
        ser = self.create_serializer(Person)

        # simple
        value = ser.to_internal_value({'name': 'Alice', 'length': 123, 'birth_date': '2020-02-02'})
        self.assertEqual(value, Person(name='Alice', length=123, birth_date=datetime.date(2020, 2, 2)))

        # unsupplied fields
        value = ser.to_internal_value({'name': 'Alice', 'length': 123})
        self.assertEqual(value, Person(name='Alice', length=123, birth_date=empty))

        # unsupplied required fields (in partial mode)
        ser = self.create_serializer(Person, arguments={'partial': True})
        value = ser.to_internal_value({'name': 'Alice'})
        self.assertEqual(value, Person(name='Alice', length=empty, birth_date=empty))

        # nested
        simple = dataclasses.make_dataclass('child', [('value', str)])
        parent = dataclasses.make_dataclass('parent', [('container', simple)])
        value = self.create_serializer(parent).to_internal_value({'container': {'value': 'a'}})
        self.assertEqual(value, parent(container=simple(value='a')))

    def test_validated_data(self):
        data = {'name': 'Alice', 'length': 123}
        ser = self.create_serializer(Person, arguments={'data': data})
        ser.is_valid(raise_exception=True)

        self.assertEqual(ser.validated_data, Person(name='Alice', length=123, birth_date=None))

    def test_hyperlinked(self):
        # test if it nests itself, as opposed to the "default" DataclassSerializer
        ser = HyperlinkedDataclassSerializer(dataclass=Group)
        f = ser.get_fields()

        self.assertIsInstance(f['leader'], HyperlinkedDataclassSerializer)
        self.assertIsInstance(f['people'].child, HyperlinkedDataclassSerializer)



# Some type definitions that can be useful.

from typing import Dict

try:
    # Python 3.8 and later
    from typing import Final, Literal, Protocol
except ImportError:
    from typing_extensions import Final, Literal, Protocol


# Note that this doesn't actually work yet (https://stackoverflow.com/a/55240861)
class Dataclass(Protocol):
    __dataclass_fields__: Dict



"""
Abstract away the runtime introspection on type hints behind an at least mostly-sane interface.

Unfortunately, Python does not really have an API for that. As such, we're mostly poking around in the internals of the
typing module and use heuristics that seem to work for common type hints. There's fairly extensive test coverage to both
document what should be supported here, and to see what needs to be fixed for a new Python version.

Requirements for this module:

  * Support for Python 3.7 - 3.9
  * Support both immediate and postponed evaluation of annotations (PEP 563, i.e. `from __future__ import annotations`)
  * Support collection generics from the typing module (i.e. typing.List[int])
  * Support standard collection generics (PEP 585, i.e. list[int])

Note that there was some promising development in the `typing_inspect` module, but it is still marked experimental and
development seems to have stalled. Maybe in the future?
"""
import collections
import typing

from .types import Final, Literal

# typing._BaseGenericAlias was split-out from GenericAlias in Python 3.9
if hasattr(typing, '_BaseGenericAlias'):
    GenericAlias = typing._BaseGenericAlias
else:
    GenericAlias = typing._GenericAlias

# Wrappers around typing.get_origin() and typing.get_args() for Python 3.7
try:
    get_origin = typing.get_origin
    get_args = typing.get_args
except AttributeError:
    def get_origin(tp: type) -> type:
        return tp.__origin__ if isinstance(tp, GenericAlias) else None

    def get_args(tp: type) -> type:
        return tp.__args__ if isinstance(tp, GenericAlias) else ()

# Some implementation notes:
# * We detect if types are a generic by whether their origin is not None.
# * Some origins (most notably literals) aren't a type, so before doing issubclass() on an origin, we need to check
#   if it's a type.


def is_iterable_type(tp: type) -> bool:
    """
    Test if the given type is iterable.

    Some examples of iterable type hints are:

        Iterable[str]
        Collection[str]
        Mapping[str, int]
        Sequence[str]
        List[str]
        Set[str]
        Dict[str, int]
        Generator[str, int, int]

    """
    # Iterables are a generic with an origin that is a subclass of Iterable.
    origin = get_origin(tp)
    return (
        origin is not None and
        isinstance(origin, type) and
        issubclass(origin, collections.abc.Iterable)
    )


def get_iterable_element_type(tp: type) -> type:
    """
    Get the type of elements in an iterable.
    """
    if not is_iterable_type(tp):
        raise ValueError('get_iterable_element_type() called with non-iterable type.')

    args = get_args(tp)
    return args[0] if len(args) > 0 else typing.Any


def is_mapping_type(tp: type) -> bool:
    """
    Test if the given type is a mapping.

    Some examples of mapping type hints are:

        Mapping[str, int]
        Dict[str, int]

    """
    # Mappings are a generic with an origin that is a subclass of Mapping.
    origin = get_origin(tp)
    return (
        origin is not None and
        isinstance(origin, type) and
        issubclass(origin, collections.abc.Mapping)
    )


def get_mapping_value_type(tp: type) -> type:
    """
    Get the type of values in a mapping.
    """
    if not is_mapping_type(tp):
        raise ValueError('get_mapping_value_type() called with non-mapping type.')

    args = get_args(tp)
    return args[1] if len(args) == 2 else typing.Any


def is_optional_type(tp: type) -> bool:
    """
    Test if the given type is optional.

    Some examples of optional type hints are:

        Optional[int]
        Union[int, None]
        Literal[0, None]

    """
    # Optional types are:
    # * a generic
    # * with an origin of typing.Union (typing.Optional[int] reduces to typing.Union[int, None] at constructor time)
    # * with at least one argument that is NoneType
    # * and at least one argumen that isn't NoneType
    # except for Literals (sigh), which are just regular Literals with None as an allowed value.
    origin = get_origin(tp)
    args = get_args(tp)
    none_type = type(None)
    return (
        origin is not None and
        origin is typing.Union and
        any(argument_type is none_type for argument_type in args) and
        any(argument_type is not none_type for argument_type in args)
    ) or (
        is_literal_type(tp) and
        None in get_literal_choices(tp)
    )


def get_optional_type(tp: type) -> type:
    """
    Get the type that is made optional.
    """
    if not is_optional_type(tp):
        raise ValueError('get_optional_type() called with non-optional type.')

    if is_literal_type(tp):
        # Note that this doesn't remove `None` as a valid choice, but that's not a problem so far.
        return tp

    return next(argument_type for argument_type in get_args(tp) if argument_type is not type(None))


def is_literal_type(tp: type) -> bool:
    """
    Test if the given type is a literal expression.
    """
    # Stolen from typing_inspect
    origin = get_origin(tp)
    return tp is Literal or origin is Literal


def get_literal_choices(tp: type) -> typing.List[typing.Union[str, bytes, int, bool, None]]:
    """
    Get the possible values for a literal type.
    """
    if not is_literal_type(tp):
        raise ValueError('get_literal_choices() called with non-literal type.')

    # A Literal type may contain other Literals, so need to unnest those. This doesn't happen automatically like it
    # happens for Unions. For more details, see
    # https://www.python.org/dev/peps/pep-0586/#legal-parameters-for-literal-at-type-check-time
    values = []
    for value in get_args(tp):
        if is_literal_type(value):
            values.extend(get_literal_choices(value))
        else:
            values.append(value)
    return values


def is_final_type(tp: type) -> bool:
    """
    Test if the given type is a final type.
    """
    return tp is Final or get_origin(tp) is Final


def get_final_type(tp: type) -> type:
    """
    Get the type that is made final.
    """
    if not is_final_type(tp):
        raise ValueError('get_final_type() called with non-final type.')

    args = get_args(tp)
    return args[0] if len(args) > 0 else typing.Any


def is_type_variable(tp: type) -> bool:
    """
    Test if the given type is a variable type.
    """
    return isinstance(tp, typing.TypeVar)


def get_variable_type_substitute(tp: typing.TypeVar) -> type:
    """
    Get the substitute for a variable type.
    """
    if len(tp.__constraints__) > 0:
        return typing.Union[tp.__constraints__]
    if tp.__bound__ is not None:
        return tp.__bound__
    return typing.Any



import typing
import unittest
import sys

from rest_framework_dataclasses import types, typing_utils


class TypingTest(unittest.TestCase):
    def assertAnyTypeEquivalent(self, tp: type):
        # In some cases we accept either typing.Any (used by Python 3.9+) or an unconstrained typevar (used by Python
        # 3.7 and 3.8). It's essentially the same, and we strip the typevar before usage anyway.
        self.assertTrue(tp is typing.Any or (isinstance(tp, typing.TypeVar) and len(tp.__constraints__) == 0))

    def test_iterable(self):
        self.assertTrue(typing_utils.is_iterable_type(typing.Iterable[str]))
        self.assertTrue(typing_utils.is_iterable_type(typing.Collection[str]))
        self.assertTrue(typing_utils.is_iterable_type(typing.Mapping[str, int]))
        self.assertTrue(typing_utils.is_iterable_type(typing.Sequence[str]))
        self.assertTrue(typing_utils.is_iterable_type(typing.List[str]))
        self.assertTrue(typing_utils.is_iterable_type(typing.Dict[str, int]))
        self.assertTrue(typing_utils.is_iterable_type(typing.Set[str]))
        self.assertTrue(typing_utils.is_iterable_type(typing.Generator[str, int, int]))
        self.assertTrue(typing_utils.is_iterable_type(typing.List))
        self.assertFalse(typing_utils.is_iterable_type(str))
        self.assertFalse(typing_utils.is_iterable_type(bytes))
        self.assertFalse(typing_utils.is_iterable_type(int))
        self.assertFalse(typing_utils.is_iterable_type(TypingTest))

        self.assertEqual(typing_utils.get_iterable_element_type(typing.Iterable[str]), str)
        self.assertEqual(typing_utils.get_iterable_element_type(typing.Collection[str]), str)
        self.assertEqual(typing_utils.get_iterable_element_type(typing.Mapping[str, int]), str)
        self.assertEqual(typing_utils.get_iterable_element_type(typing.Sequence[str]), str)
        self.assertEqual(typing_utils.get_iterable_element_type(typing.List[str]), str)
        self.assertEqual(typing_utils.get_iterable_element_type(typing.Dict[str, int]), str)
        self.assertEqual(typing_utils.get_iterable_element_type(typing.Set[str]), str)
        self.assertEqual(typing_utils.get_iterable_element_type(typing.Generator[str, int, int]), str)
        self.assertAnyTypeEquivalent(typing_utils.get_iterable_element_type(typing.List))

        with self.assertRaises(ValueError):
            typing_utils.get_iterable_element_type(str)

    def test_mapping(self):
        self.assertTrue(typing_utils.is_mapping_type(typing.Mapping[str, int]))
        self.assertTrue(typing_utils.is_mapping_type(typing.Dict[str, int]))
        self.assertTrue(typing_utils.is_mapping_type(typing.Dict))
        self.assertFalse(typing_utils.is_mapping_type(str))
        self.assertFalse(typing_utils.is_mapping_type(int))
        self.assertFalse(typing_utils.is_mapping_type(TypingTest))

        self.assertEqual(typing_utils.get_mapping_value_type(typing.Mapping[str, int]), int)
        self.assertEqual(typing_utils.get_mapping_value_type(typing.Dict[str, int]), int)
        self.assertAnyTypeEquivalent(typing_utils.get_mapping_value_type(typing.Dict))

        with self.assertRaises(ValueError):
            typing_utils.get_mapping_value_type(str)

    @unittest.skipIf(sys.version_info < (3, 9, 0), 'Python 3.9 required')
    def test_pep585(self):
        self.assertTrue(typing_utils.is_iterable_type(list[int]))
        self.assertEqual(typing_utils.get_iterable_element_type(list[int]), int)

        self.assertTrue(typing_utils.is_mapping_type(dict[str, int]))
        self.assertTrue(typing_utils.get_mapping_value_type(dict[str, int]), int)

    def test_optional(self):
        self.assertTrue(typing_utils.is_optional_type(typing.Optional[str]))
        self.assertTrue(typing_utils.is_optional_type(typing.Union[str, None]))
        self.assertTrue(typing_utils.is_optional_type(typing.Union[str, typing.Optional[str]]))
        self.assertTrue(typing_utils.is_optional_type(typing.Union[str, typing.Union[str, None]]))
        self.assertFalse(typing_utils.is_optional_type(str))
        self.assertFalse(typing_utils.is_optional_type(int))
        self.assertFalse(typing_utils.is_optional_type(TypingTest))

        self.assertEqual(typing_utils.get_optional_type(typing.Optional[str]), str)
        self.assertEqual(typing_utils.get_optional_type(typing.Union[str, None]), str)
        self.assertEqual(typing_utils.get_optional_type(typing.Union[str, typing.Optional[str]]), str)
        self.assertEqual(typing_utils.get_optional_type(typing.Union[str, typing.Union[str, None]]), str)

        with self.assertRaises(ValueError):
            typing_utils.get_optional_type(str)

    def test_final(self):
        self.assertTrue(typing_utils.is_final_type(types.Final[int]))
        self.assertTrue(typing_utils.is_final_type(types.Final))
        self.assertFalse(typing_utils.is_final_type(int))

        self.assertEqual(typing_utils.get_final_type(types.Final[int]), int)
        self.assertAnyTypeEquivalent(typing_utils.get_final_type(types.Final))

        with self.assertRaises(ValueError):
            typing_utils.get_final_type(str)

    def test_literal(self):
        self.assertTrue(typing_utils.is_literal_type(types.Literal['a', 'b']))
        self.assertTrue(typing_utils.is_literal_type(types.Literal['a', 'b', None]))
        self.assertTrue(typing_utils.is_literal_type(types.Literal['a', 'b', types.Literal['c', 'd']]))
        self.assertTrue(typing_utils.is_literal_type(types.Literal['a', 'b', types.Literal['c', 'd', None]]))
        self.assertTrue(typing_utils.is_literal_type(types.Literal['a', 'b', types.Literal[1, 2]]))
        self.assertTrue(typing_utils.is_literal_type(types.Literal['a', 'b', types.Literal[1, 2, None]]))

        self.assertFalse(typing_utils.is_optional_type(types.Literal['a', 'b']))
        self.assertTrue(typing_utils.is_optional_type(types.Literal['a', 'b', None]))
        self.assertFalse(typing_utils.is_optional_type(types.Literal['a', 'b', types.Literal['c', 'd']]))
        self.assertTrue(typing_utils.is_optional_type(types.Literal['a', 'b', types.Literal['c', 'd', None]]))
        self.assertFalse(typing_utils.is_optional_type(types.Literal['a', 'b', types.Literal[1, 2]]))
        self.assertTrue(typing_utils.is_optional_type(types.Literal['a', 'b', types.Literal[1, 2, None]]))

        self.assertListEqual(typing_utils.get_literal_choices(types.Literal['a', 'b']), ['a', 'b'])
        self.assertListEqual(typing_utils.get_literal_choices(types.Literal['a', 'b', None]), ['a', 'b', None])
        self.assertListEqual(typing_utils.get_literal_choices(types.Literal['a', 'b', types.Literal['c', 'd']]),
                             ['a', 'b', 'c', 'd'])
        self.assertListEqual(typing_utils.get_literal_choices(types.Literal['a', 'b', types.Literal['c', 'd', None]]),
                             ['a', 'b', 'c', 'd', None])
        self.assertListEqual(typing_utils.get_literal_choices(types.Literal['a', 'b', types.Literal[1, 2]]),
                             ['a', 'b', 1, 2])
        self.assertListEqual(typing_utils.get_literal_choices(types.Literal['a', 'b', types.Literal[1, 2, None]]),
                             ['a', 'b', 1, 2, None])

        with self.assertRaises(ValueError):
            typing_utils.get_literal_choices(str)

    # noinspection PyPep8Naming
    def test_variable_type(self):
        T = typing.TypeVar('T')
        U = typing.TypeVar('U', int, str)
        V = typing.TypeVar('V', bound=Exception)

        self.assertTrue(typing_utils.is_type_variable(T))
        self.assertTrue(typing_utils.is_type_variable(U))
        self.assertTrue(typing_utils.is_type_variable(V))
        self.assertFalse(typing_utils.is_type_variable(int))
        self.assertFalse(typing_utils.is_type_variable(typing.List))

        self.assertEqual(typing_utils.get_variable_type_substitute(T), typing.Any)
        self.assertEqual(typing_utils.get_variable_type_substitute(U), typing.Union[int, str])
        self.assertEqual(typing_utils.get_variable_type_substitute(V), Exception)



import dataclasses
import typing

from unittest import TestCase

from rest_framework import fields
from rest_framework_dataclasses.serializers import DataclassSerializer


@dataclasses.dataclass
class Simple:
    value: str


class SimpleSerializer(DataclassSerializer):
    class Meta:
        dataclass = Simple


class IssuesTest(TestCase):
    def check_deserialize(self, serializer=DataclassSerializer, **kwargs):
        serializer = serializer(**kwargs)
        serializer.is_valid(raise_exception=True)
        serializer.save()

    # Issue #3: save() with nested dataclasses is broken
    def test_save_nested_dataclass(self):
        parent = dataclasses.make_dataclass('Parent', [('nested', Simple)])
        data = {'nested': {'value': 'A'}}

        self.check_deserialize(dataclass=parent, data=data)

    # Issue #13: save() with nested list items is broken
    def test_nested_list(self):
        parent = dataclasses.make_dataclass('Parent', [('nested', typing.List[Simple])])
        data = {'nested': [{'value': 'A'}]}

        class ParentSerializer(DataclassSerializer):
            nested = SimpleSerializer(many=True)

        self.check_deserialize(ParentSerializer, dataclass=parent, data=data)

    # Issue #15: save() with nullable nested is broken
    def test_nested_nullable(self):
        parent = dataclasses.make_dataclass('Parent', [('nested', typing.Optional[Simple])])
        data = {'nested': None}

        class ParentSerializer(DataclassSerializer):
            nested = SimpleSerializer(allow_null=True, required=False)

        self.check_deserialize(ParentSerializer, dataclass=parent, data=data)

    # Issue #19: create() breaks with source parameter
    def test_create_source(self):
        data = {'renamed_value': 'a'}

        class SimpleRenamedSerializer(DataclassSerializer):
            renamed_value = fields.CharField(source='value')

            class Meta:
                dataclass = Simple
                fields = ('renamed_value', )

        self.check_deserialize(SimpleRenamedSerializer, data=data)

    # Issue #39: many=True results in empty sentinel in validated_data for optional fields
    def test_many_empty(self):
        @dataclasses.dataclass
        class HelloWorld:
            value: str = 'default'

        serializer = DataclassSerializer(dataclass=HelloWorld, many=True, data=[{}])
        serializer.is_valid(raise_exception=True)
        data = serializer.validated_data

        self.assertIsNot(fields.empty, data[0].value)
        self.assertEqual('default', data[0].value)



import copy
import dataclasses
import datetime
import decimal
import uuid
from collections import OrderedDict
from enum import Enum
from typing import Any, Dict, Generic, Iterable, Mapping, Tuple, Type, TypeVar

import rest_framework.fields
import rest_framework.serializers
from django.core.exceptions import ImproperlyConfigured
from django.db.models import Model
from django.utils.functional import cached_property
from rest_framework.fields import empty
from rest_framework.relations import HyperlinkedRelatedField, PrimaryKeyRelatedField
from rest_framework.utils.field_mapping import get_relation_kwargs

from rest_framework_dataclasses import fields, field_utils, typing_utils
from rest_framework_dataclasses.field_utils import get_dataclass_definition, DataclassDefinition, TypeInfo
from rest_framework_dataclasses.types import Dataclass


# Define some types to make type hinting more readable
KWArgs = Dict[str, Any]
SerializerField = rest_framework.fields.Field
SerializerFieldDefinition = Tuple[Type[SerializerField], KWArgs]
T = TypeVar('T', bound=Dataclass)
AnyT = TypeVar('AnyT')


# Helper function to strip the empty sentinel value and replace it with the default value from a dataclass
def _strip_empty_sentinels(data: AnyT, instance: AnyT = None) -> AnyT:
    if dataclasses.is_dataclass(data) and not isinstance(data, type):
        values = {field.name: _strip_empty_sentinels(getattr(data, field.name),
                                                     getattr(instance, field.name, None))
                  for field in dataclasses.fields(data)
                  if getattr(data, field.name) is not empty}
        if instance:
            for field, value in values.items():
                setattr(instance, field, value)
            return instance
        else:
            return type(data)(**values)
    elif isinstance(data, list):
        return [_strip_empty_sentinels(item) for item in data]
    elif isinstance(data, dict):
        return {key: _strip_empty_sentinels(value) for key, value in data.items()}
    return data


# noinspection PyMethodMayBeStatic
class DataclassSerializer(rest_framework.serializers.Serializer, Generic[T]):
    """
    A `DataclassSerializer` is just a regular `Serializer`, except that:

    * A set of default fields are automatically populated.
    * A set of default validators are automatically populated.
    * Default `.create()` and `.update()` implementations are provided.

    The process of automatically determining a set of serializer fields based on the dataclass fields is slightly
    complex, but you almost certainly don't need to dig into the implementation.

    If the `DataclassSerializer` class *doesn't* generate the set of fields that you need you should either declare the
    extra/differing fields explicitly on the serializer class, or simply use a `Serializer` class.
    """

    # The mapping of field types to serializer fields
    serializer_field_mapping = {
        int:                rest_framework.fields.IntegerField,
        float:              rest_framework.fields.FloatField,
        bool:               rest_framework.fields.BooleanField,
        str:                rest_framework.fields.CharField,
        decimal.Decimal:    fields.DefaultDecimalField,
        datetime.date:      rest_framework.fields.DateField,
        datetime.datetime:  rest_framework.fields.DateTimeField,
        datetime.time:      rest_framework.fields.TimeField,
        datetime.timedelta: rest_framework.fields.DurationField,
        uuid.UUID:          rest_framework.fields.UUIDField,
        dict:               rest_framework.fields.DictField,
        list:               rest_framework.fields.ListField
    }
    serializer_related_field = PrimaryKeyRelatedField

    # Unfortunately this cannot be an actual field as Python processes the class before it defines the class, but this
    # comes close enough.
    @property
    def serializer_dataclass_field(self):
        return DataclassSerializer

    # Type hints
    _declared_fields: Mapping[str, SerializerField]

    # Override constructor to allow "anonymous" usage by passing the dataclass type and extra kwargs as a constructor
    # parameter instead of via a Meta class.
    def __init__(self, *args, **kwargs):
        self.dataclass = kwargs.pop('dataclass', None)
        self.extra_kwargs = kwargs.pop('extra_kwargs', None)
        super(DataclassSerializer, self).__init__(*args, **kwargs)

    @classmethod
    def many_init(cls, *args, **kwargs):
        """
        Implements the creation of a `DataclassListSerializer` parent class when `many=True` is used.
        """
        list_kwargs = {key: value for key, value in kwargs.items()
                       if key in rest_framework.serializers.LIST_SERIALIZER_KWARGS}
        list_kwargs['child'] = cls(*args, **kwargs)

        allow_empty = kwargs.pop('allow_empty', None)
        if allow_empty is not None:
            list_kwargs['allow_empty'] = allow_empty

        meta = getattr(cls, 'Meta', None)
        list_serializer_class = getattr(meta, 'list_serializer_class', DataclassListSerializer)
        return list_serializer_class(*args, **list_kwargs)

    # Parse and validate the configuration to a more usable format

    @cached_property
    def dataclass_definition(self) -> DataclassDefinition:
        # Determine the dataclass that we should operate on.
        if self.dataclass:
            assert not hasattr(self, 'Meta'), (
                "Class '{serializer_class}' may not have `Meta` attribute when instantiated with `dataclass` parameter."
                .format(serializer_class=self.__class__.__name__)
            )

            dataclass_type = self.dataclass
        else:
            assert hasattr(self, 'Meta'), (
                "Class '{serializer_class}' missing `Meta` attribute."
                .format(serializer_class=self.__class__.__name__)
            )

            meta = getattr(self, 'Meta')
            assert hasattr(meta, 'dataclass'), (
                "Class '{serializer_class}' missing `Meta.dataclass` attribute."
                .format(serializer_class=self.__class__.__name__)
            )

            dataclass_type = meta.dataclass

        # Make sure we're dealing with an actual dataclass.
        if not dataclasses.is_dataclass(dataclass_type):
            raise ValueError(
                "Class '{serializer_class}' can only be used to serialize dataclasses."
                .format(serializer_class=self.__class__.__name__)
            )

        return get_dataclass_definition(dataclass_type)

    # Default `create` and `update` behavior...

    def create(self, validated_data: T) -> T:
        return _strip_empty_sentinels(validated_data)

    def update(self, instance: T, validated_data: T) -> T:
        return _strip_empty_sentinels(validated_data, instance)

    def save(self, **kwargs) -> T:
        assert hasattr(self, '_errors'), (
            "You must call `.is_valid()` before calling `.save()`."
        )

        assert not self.errors, (
            "You cannot call `.save()` on a serializer with invalid data."
        )

        assert not hasattr(self, '_data'), (
            "You cannot call `.save()` after accessing `serializer.data`."
            "If you need to access data before committing to the dataclass then "
            "inspect 'serializer.validated_data' instead. "
        )

        # Explicitly use internal validated_data here, as we want the empty sentinel values instead of the normalized
        # external representation.
        validated_data = dataclasses.replace(self._validated_data, **kwargs)

        if self.instance is not None:
            self.instance = self.update(self.instance, validated_data)
        else:
            self.instance = self.create(validated_data)

        assert self.instance is not None, (
            '`update()` or `create()` did not return an object instance.'
        )

        return self.instance

    # Determine the fields to apply...

    def get_fields(self) -> Dict[str, SerializerField]:
        """
        Return the dict of field names -> field instances that should be used for `self.fields` when instantiating the
        serializer.
        """
        # Copy declared fields, so that the field on the serializer class remains unchanged when the fields are bound to
        # the serializer instance.
        declared_fields = copy.deepcopy(self._declared_fields)

        # Copy fields declared in metadata, for the same reason.
        metadata_fields = {field.name: copy.deepcopy(field.metadata['serializer_field'])
                           for field in self.dataclass_definition.fields.values()
                           if 'serializer_field' in field.metadata}

        # Determine all fields that should be included on the serializer.
        field_names = self.get_field_names()

        # Determine any extra field arguments that should be included.
        extra_kwargs = self.get_extra_kwargs()

        # Determine the fields that should be included on the serializer.
        fields = OrderedDict()

        for field_name in field_names:
            # If the field is explicitly declared, either on the serializer or in the field metadata, then use that.
            if field_name in declared_fields or field_name in metadata_fields:
                assert field_name not in extra_kwargs, (
                    "Cannot both declare the field '{field_name}' and include it in '{serializer_class}' "
                    "`extra_kwargs` or `read_only_fields` option. Move all options to the field declaration."
                    .format(field_name=field_name, serializer_class=self.__class__.__name__)
                )

                # Prefer the field declared on the serializer above the one declared in field metadata.
                fields[field_name] = declared_fields.get(field_name, metadata_fields.get(field_name, None))
                continue

            extra_field_kwargs = extra_kwargs.get(field_name, {})
            fields[field_name] = self.create_field(field_name, extra_field_kwargs)

        return fields

    # Methods for determining the set of field names to include...

    def get_field_names(self) -> Iterable[str]:
        """
        Returns the list of all field names that should be created when instantiating this serializer class. This is
        based on the default set of fields, but also takes into account the `Meta.fields` or `Meta.exclude` options
        if they have been specified.
        """
        # Retrieve names of the declared fields.
        declared_field_names = self._declared_fields.keys()

        # Read configuration from Meta class.
        meta = getattr(self, 'Meta', None)
        fields = getattr(meta, 'fields', None)
        exclude = getattr(meta, 'exclude', None)

        if fields and fields != rest_framework.serializers.ALL_FIELDS and not isinstance(fields, (list, tuple)):
            raise TypeError(
                "The `fields` option must be a list or tuple or '__all__'. Got '{type}'."
                .format(type=type(fields).__name__)
            )

        if exclude and not isinstance(exclude, (list, tuple)):
            raise TypeError(
                "The `exclude` option must be a list or tuple. Got '{type}'."
                .format(type=type(exclude).__name__)
            )

        assert not (fields and exclude), (
            "Cannot set both `fields` and `exclude` options on serializer '{serializer_class}'."
            .format(serializer_class=self.__class__.__name__)
        )

        # If fields is not specified, or is the magic all fields option, make it a list consisting of the all fields
        # placeholder for now.
        if fields is None or fields == rest_framework.serializers.ALL_FIELDS:
            fields = [rest_framework.serializers.ALL_FIELDS]

        # For explicitly specified fields, ensure that they are valid.
        for field_name in fields:
            assert (
                field_name == rest_framework.serializers.ALL_FIELDS or                         # all fields magic option
                field_name in self.dataclass_definition.fields or                              # dataclass fields
                field_name in declared_field_names or                                          # declared fields
                callable(getattr(self.dataclass_definition.dataclass_type, field_name, None))  # methods
            ), (
                "The field '{field_name}' was included on serializer {serializer_class} in the `fields` option, "
                "but does not match any dataclass field."
                .format(field_name=field_name, serializer_class=self.__class__.__name__)
            )

        if rest_framework.serializers.ALL_FIELDS not in fields:
            # If there are only explicitly specified fields, ensure that all declared fields are included. Do not
            # require any fields that are declared in a parent class, in order to allow serializer subclasses to only
            # include a subset of fields.
            required_field_names = set(declared_field_names)
            for cls in self.__class__.__bases__:
                required_field_names -= set(getattr(cls, '_declared_fields', []))

            for field_name in required_field_names:
                assert field_name in fields, (
                    "The field '{field_name}' was declared on serializer '{serializer_class}', but has not been "
                    "included in the `fields` option."
                    .format(field_name=field_name, serializer_class=self.__class__.__name__)
                )
            return list(fields)

        # The field list now includes the magic all fields option, so replace it with the default field names.
        fields = list(fields)
        fields.remove(rest_framework.serializers.ALL_FIELDS)
        fields.extend(self.get_default_field_names(declared_field_names))

        if exclude is not None:
            # If `Meta.exclude` is included, then remove those fields.
            for field_name in exclude:
                assert field_name not in declared_field_names, (
                    "Cannot both declare the field '{field_name}' and include it in '{serializer_class}' `exclude` "
                    "option. Remove the field or, if inherited from a parent serializer, disable with "
                    "`{field_name} = None`."
                    .format(field_name=field_name, serializer_class=self.__class__.__name__)
                )

                assert field_name in fields, (
                    "The field '{field_name}' was included on serializer {serializer_class} in the `exclude` option, "
                    "but does not match any dataclass field."
                    .format(field_name=field_name, serializer_class=self.__class__.__name__)
                )

                fields.remove(field_name)

        return fields

    def get_default_field_names(self, declared_fields: Iterable[str]) -> Iterable[str]:
        """
        Return the default list of field names that will be used if the `Meta.fields` option is not specified.
        """
        return (
            list(declared_fields) +
            list(self.dataclass_definition.fields)
        )

    # Methods for constructing serializer fields...

    def create_field(self, field_name: str, extra_kwargs: KWArgs) -> SerializerField:
        # Get the source field (this can be useful to represent a single dataclass field in multiple ways in the
        # serialization state).
        source = extra_kwargs.get('source', '*')
        if source == '*':
            source = field_name

        # Determine the serializer field class and keyword arguments.
        if source in self.dataclass_definition.fields:
            field = self.dataclass_definition.fields[source]
            type_info = field_utils.get_type_info(self.dataclass_definition.field_types[source])
            field_class, field_kwargs = self.build_typed_field(source, type_info, extra_kwargs)

            # Include extra kwargs defined in the field metadata
            if 'serializer_kwargs' in field.metadata:
                field_kwargs = self.include_extra_kwargs(field_kwargs, field.metadata['serializer_kwargs'])
        elif hasattr(self.dataclass_definition.dataclass_type, source):
            field_class, field_kwargs = self.build_property_field(source)
        else:
            field_class, field_kwargs = self.build_unknown_field(source)

        # Include any extra kwargs defined through `Meta.extra_kwargs`
        field_kwargs = self.include_extra_kwargs(field_kwargs, extra_kwargs)

        # Create the serializer field instance.
        return field_class(**field_kwargs)

    def build_typed_field(self, field_name: str, type_info: TypeInfo,
                          extra_kwargs: KWArgs) -> SerializerFieldDefinition:
        """
        Create a serializer field for a typed dataclass field.
        """
        if type_info.is_mapping or type_info.is_many:
            field_class, field_kwargs = self.build_composite_field(field_name, type_info, extra_kwargs)
        elif dataclasses.is_dataclass(type_info.base_type):
            field_class, field_kwargs = self.build_dataclass_field(field_name, type_info)
        elif isinstance(type_info.base_type, type) and issubclass(type_info.base_type, Model):
            field_class, field_kwargs = self.build_relational_field(field_name, type_info)
        elif isinstance(type_info.base_type, type) and issubclass(type_info.base_type, Enum):
            field_class, field_kwargs = self.build_enum_field(field_name, type_info)
        elif typing_utils.is_literal_type(type_info.base_type):
            field_class, field_kwargs = self.build_literal_field(field_name, type_info)
        else:
            field_class, field_kwargs = self.build_standard_field(field_name, type_info)

        # Mark a field as not-required if it has a default value (factory) on the dataclass. This is consistent with the
        # constructor of dataclasses, where these fields are also made optional. Note that this is different from the
        # `typing.Optional[]` qualifier, which merely makes the field nullable, but still requires it to be passed. Of
        # course it makes sense for `Optional` fields to have `None` as a default value, but that's up to the user.
        field = self.dataclass_definition.fields[field_name]
        if field.default is not dataclasses.MISSING or field.default_factory is not dataclasses.MISSING:
            field_kwargs['required'] = False

            # Explicitly don't set the default argument here. Setting it would cause the default value to be inserted in
            # the native representation (`to_internal_value()` argument) if the field wasn't supplied by the user (for
            # non-partial updates). This in turn would cause `update()` to overwrite non-supplied fields with the
            # defaults, which is undesirable. Instead, let the dataclass constructor apply the default values when the
            # dataclass is instantiated.

        # Mark a field as nullable if it is declared as Optional[] (which has a confusing name).
        if type_info.is_nullable:
            field_kwargs['allow_null'] = True

        # The final qualifier declares that a variable or attribute should not be reassigned (PEP 591). Mark the field
        # as read only.
        if type_info.is_final:
            field_kwargs['read_only'] = True

        return field_class, field_kwargs

    def build_composite_field(self, field_name: str, type_info: TypeInfo,
                              extra_kwargs: KWArgs) -> SerializerFieldDefinition:
        """
        Create a composite (mapping or list) field.
        """
        # Lookup the types from the field mapping, so that it can easily be changed without overriding the method.
        if type_info.is_mapping:
            field_class = self.serializer_field_mapping[dict]
        else:
            field_class = self.serializer_field_mapping[list]

        # If the base type is not specified or is any type, we don't have to bother creating the child field.
        if type_info.base_type is None:
            return field_class, {}

        # Recurse to build the child field (i.e. the field of every instance). We pass the extra kwargs that are
        # specified for the child field through, so these can be used to recursively specify kwargs for child fields.
        extra_child_field_kwargs = extra_kwargs.get('child_kwargs', {})
        base_type_info = field_utils.get_type_info(type_info.base_type)
        child_field_class, child_field_kwargs = self.build_typed_field(field_name, base_type_info,
                                                                       extra_child_field_kwargs)

        # Include the extra kwargs specified for the child field before instantiating it.
        child_field_kwargs = self.include_extra_kwargs(child_field_kwargs, extra_child_field_kwargs)

        # Create child field and initialize parent field kwargs
        child_field = child_field_class(**child_field_kwargs)
        field_kwargs = {'child': child_field}
        return field_class, field_kwargs

    def build_standard_field(self, field_name: str, type_info: TypeInfo) -> SerializerFieldDefinition:
        """
        Create regular dataclass fields.
        """
        try:
            field_class = field_utils.lookup_type_in_mapping(self.serializer_field_mapping, type_info.base_type)
            field_kwargs = {}

            return field_class, field_kwargs
        except KeyError:
            # When resolving the type hint fails, raise a nice descriptive error based on the outermost type of the
            # field (this makes solving deep recursive errors much easier).
            field_type = self.dataclass_definition.field_types[field_name]
            raise NotImplementedError(
                "Automatic serializer field deduction not supported for field '{field}' on '{dataclass}' "
                "of type '{type}' (during search for field of type '{reduced_type}')."
                .format(dataclass=self.dataclass_definition.dataclass_type.__name__, field=field_name,
                        type=field_type, reduced_type=type_info.base_type)
            )

    def build_relational_field(self, field_name: str, type_info: TypeInfo) -> SerializerFieldDefinition:
        """
        Create fields for models.
        """
        relation_info = field_utils.get_relation_info(type_info)
        field_class = self.serializer_related_field
        field_kwargs = get_relation_kwargs(field_name, relation_info)

        # `view_name` is only valid for hyperlinked relationships.
        if not issubclass(field_class, HyperlinkedRelatedField):
            field_kwargs.pop('view_name', None)

        return field_class, field_kwargs

    def build_enum_field(self, field_name: str, type_info: TypeInfo) -> SerializerFieldDefinition:
        """
        Create EnumField from a Enum type.
        """
        field_class = fields.EnumField
        field_kwargs = {
            'enum_class': type_info.base_type
        }
        return field_class, field_kwargs

    # noinspection PyUnusedLocal
    def build_literal_field(self, field_name: str, type_info: TypeInfo) -> SerializerFieldDefinition:
        """
        Create ChoiceField from a Literal[...] type.
        """
        field_class = rest_framework.fields.ChoiceField
        choices = typing_utils.get_literal_choices(type_info.base_type)
        field_kwargs = {
            'choices': [val for val in choices if val not in (None, '')],
            'allow_blank': '' in choices,
        }

        return field_class, field_kwargs

    def build_dataclass_field(self, field_name: str, type_info: TypeInfo) -> SerializerFieldDefinition:
        """
        Create fields for dataclasses.
        """
        try:
            field_class = field_utils.lookup_type_in_mapping(self.serializer_field_mapping, type_info.base_type)
        except KeyError:
            field_class = self.serializer_dataclass_field

        field_kwargs = {'dataclass': type_info.base_type,
                        'many': type_info.is_many}

        return field_class, field_kwargs

    # noinspection PyUnusedLocal
    def build_property_field(self, field_name: str) -> SerializerFieldDefinition:
        """
        Create a read only field for dataclass methods and properties.
        """
        field_class = rest_framework.fields.ReadOnlyField
        field_kwargs = {}

        return field_class, field_kwargs

    def build_unknown_field(self, field_name: str) -> SerializerFieldDefinition:
        """
        Raise an error on any unknown fields.
        """
        raise ImproperlyConfigured(
            "Field name '{field_name}' is not valid for dataclass '{class_name}'."
            .format(field_name=field_name, class_name=self.dataclass_definition.dataclass_type.__name__)
        )

    # Methods for determining additional keyword arguments to apply...

    def include_extra_kwargs(self, kwargs: KWArgs, extra_kwargs: KWArgs) -> KWArgs:
        """
        Include any `extra_kwargs` that have been included for this field, possibly removing any incompatible existing
        keyword arguments.
        """
        # If the field is made read only, drop write-only related arguments.
        if extra_kwargs.get('read_only', False):
            for attr in [
                'required', 'default', 'allow_blank', 'allow_null',
                'min_length', 'max_length', 'min_value', 'max_value',
                'validators', 'queryset'
            ]:
                kwargs.pop(attr, None)

        # If `default` is specified, `required` may not be specified. If the field is not explicitly specified as not
        # required (the behaviour of `default`), remove that specification.
        if extra_kwargs.get('default') and kwargs.get('required') is False:
            kwargs.pop('required')

        if extra_kwargs.get('read_only', kwargs.get('read_only', False)):
            extra_kwargs.pop('required', None)  # Read only fields should always omit the 'required' argument.

        if 'child_kwargs' in extra_kwargs:
            extra_kwargs.pop('child_kwargs', None)  # Always drop the child_kwargs field, as it's applied manually

        kwargs.update(extra_kwargs)

        return kwargs

    def get_extra_kwargs(self) -> KWArgs:
        """
        Return a dictionary mapping field names to a dictionary of additional keyword arguments.
        """
        meta = getattr(self, 'Meta', None)
        extra_kwargs = copy.deepcopy(getattr(meta, 'extra_kwargs', None)) if meta is not None else None
        if extra_kwargs is None:
            extra_kwargs = self.extra_kwargs or {}

        read_only_fields = getattr(meta, 'read_only_fields', None)
        if read_only_fields is not None:
            if not isinstance(read_only_fields, (list, tuple)):
                raise TypeError(
                    "The `read_only_fields` option must be a list or tuple. Got '{type}'."
                    .format(type=type(read_only_fields).__name__)
                )

            for field_name in read_only_fields:
                kwargs = extra_kwargs.get(field_name, {})
                kwargs['read_only'] = True
                extra_kwargs[field_name] = kwargs
        else:
            # Guard against the possible misspelling `readonly_fields` (used by the Django admin and others).
            assert not hasattr(meta, 'readonly_fields'), (
                "Serializer '{serializer_class}' has field `readonly_fields`; the correct spelling for the option is "
                "`read_only_fields`."
                .format(serializer_class=self.__class__.__name__)
            )

        return extra_kwargs

    # Methods to convert between internal normalized value and serialized representation.

    def to_internal_value(self, data: Dict[str, Any]) -> T:
        """
        Convert a dictionary representation of the dataclass containing only primitive values to a dataclass instance.
        """
        native_values = super(DataclassSerializer, self).to_internal_value(data)
        empty_values = {key: empty for key in self.dataclass_definition.fields.keys() if key not in native_values}

        dataclass_type = self.dataclass_definition.dataclass_type
        instance = dataclass_type(**native_values, **empty_values)

        return instance

    @cached_property
    def validated_data(self):
        """
        Replace empty sentinel value with default value in public representation. Note that this doesn't work for
        partial updates.
        """
        internal_validated_data = super(DataclassSerializer, self).validated_data
        return _strip_empty_sentinels(internal_validated_data)


class DataclassListSerializer(rest_framework.serializers.ListSerializer):
    @cached_property
    def validated_data(self):
        """
        Replace empty sentinel value with default value in public representation.
        """
        internal_validated_data = super(DataclassListSerializer, self).validated_data
        return _strip_empty_sentinels(internal_validated_data)


class HyperlinkedDataclassSerializer(DataclassSerializer):
    serializer_related_field = HyperlinkedRelatedField

    @property
    def serializer_dataclass_field(self):
        return HyperlinkedDataclassSerializer
